import matplotlib.pyplot as plt
import numpy as np


# Funktion zur Berechnung der 2-Norm
def norm(vec):
    sum_of_squares = 0
    for i in range(0, len(vec)):
        sum_of_squares = sum_of_squares + vec[i] ** 2

    return np.sqrt(sum_of_squares)


# Test, ob die Matrix M positiv definit ist, mittels Cholesky-Zerlegung
def ist_positiv_definit(m):
    try:
        np.linalg.cholesky(m)
        return True
    except np.linalg.LinAlgError:
        return False


n = 1000
h = 1 / (n - 1)

# Initialisierung der Matrix A und des Vektor f für LGS Au = f
A = np.diag(-1 * np.ones(n - 1), k=1) + np.diag(2 * np.ones(n), k=0) + np.diag(-1 * np.ones(n - 1), k=-1)
f = h ** 2 * 2 * np.ones(n)

# Teste, ob A positiv definitv ist mittels Eigenwerten
if np.all(np.linalg.eigvals(A) <= 0):  # Prüfen, ob alle Eigenwerte positiv sind
    raise ValueError("A ist nicht positiv definit.")

# Teste, ob A positiv definit ist mittels Cholesky-Zerlegung
if not (ist_positiv_definit(A)):
    raise ValueError("A ist nicht positiv definit.")

# Teste, ob A symmetrisch ist
if (A != A.T).all():
    raise ValueError("A ist nicht symmetrisch.")

# Intialisierung des Startvektors x
x = np.ones(n)

# Toleranz epsilon
eps = 0.001

count = 1

# Anfangswerte berechnen
r = f - A @ x  # Anfangsresiduum
p = r  # Anfangsabstiegsrichtung

print("0. Iterationsschritt: \n")
print("Startvektor:", x)
print("Norm des Anfangsresiduums: ", norm(r))

while eps < norm(r) < 1000:
    z = A @ p  # Matrix-Vektorprodukt berechnen und speichern

    alpha = (np.dot(r, p)) / (np.dot(p, z))
    print(alpha)

    x = x + alpha * p  # neue Itterierte x
    r = r - alpha * z  # neues Residuum

    # Bestimmung der neuen Suchrichtung
    beta = - (np.dot(r, z)) / (np.dot(p, z))
    p = r + beta * p  # neue konjugierte Abstiegsrichtung

    print(count, ". Iterationsschritt: \n")
    print("Aktuelle Iterierte:", x)
    print("Norm des Residuums: ", norm(r))

    count = count + 1

# Vergleich mit numpy-interner Lsg
u = np.linalg.solve(A, f)

print("Lösung mit CG-Verfahren:", x)
print("Numpy interne Lösung:", u)

if norm(u - x) > eps:
    print("Der CG-Algorithmus hat nicht richtig funktioniert!")
else:
    print("Der CG-Algorithmus war erfolgreich.")

plt.plot(x, linewidth=2)
plt.plot(u, linewidth=2)

plt.show()
